"""
discord_ui.slash
~~~~~~~~~~~~~~~~~~~

The slash module for this libary

- - -

Here is everything about slashcommmands.
You shouldn't need to import this module unless you really now what you're doing. 

The only thing you might wanna import is `.ext`, but you can import that from `discord_ui.ext`.
Everything else here could only be needed for intellisense.

- - -

[Github](https://github.com/discord-py-ui/discord-ui/tree/main/discord_ui/slash)
"""

from . import ext
from .types import SlashCommand, SlashSubcommand, SlashOption, ContextCommand, MessageCommand, UserCommand, SlashPermission, OptionType, CommandType